#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <arpa/inet.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/socket.h>

#include <netdb.h>
#include <net/if.h>
#include <net/if_arp.h>
#include <ifaddrs.h>

#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include <adwin-macros.h>
#include "adconfig.h"
#include "configfile.h"


#define CONFIG		"/proc/adwin/config"

extern int has_ip;

int add(void) {
	switch(add_dev.dev_type) {
		case DC_LINK:
			if ((signed)add_dev.dev_link_base_addr < 0) {
				fprintf(stderr, "error: link base address missing\n");
				return 1;
			}
			break;

		case DC_USB:
			if (add_dev.dev_usb_serial < 0) {
				fprintf(stderr, "error: usb serial missing\n");
				return 1;
			}
			break;
		case DC_ETHERNET:
			if (! has_ip) {
				fprintf(stderr, "error: host IP missing\n");
				return 1;
			}
			break;
		default:
			fprintf(stderr, "error: no device type specified\n");
			return 1;
	}
#ifdef KERNEL_MODULE
	int fd = open(CONFIG, O_WRONLY);
	if (fd < 0) {
		fprintf(stderr, "error: failed to open %s (%s)\n", CONFIG, strerror(errno));
		return 1;
	}
	if (ioctl(fd, ADIOKDEVADD, &add_dev) < 0) {
		fprintf(stderr, "error: ADIOKDEVADD failed (%s)\n", strerror(errno));
		close(fd);
		return 1;
	}
	close(fd);
#else // no KERNEL_MODULE
	addDeviceCF(add_dev);
#endif // KERNEL_MODULE

	return 0;
}

int del(void) {
#ifdef KERNEL_MODULE
	int fd;
	fd = open(CONFIG, O_WRONLY);
	if (fd < 0) {
		fprintf(stderr, "error: failed to open %s (%s)\n", CONFIG, strerror(errno));
		return 1;
	}
	if (ioctl(fd, ADIOKDEVDEL, del_dev) < 0) {
		fprintf(stderr, "error: ADIOKDEVDEL failed (%s)\n", strerror(errno));
		close(fd);
		return 1;
	}
	close(fd);
#else 
	PRINTDEBUG(DBG_CONFIGFILE, ("calling deleteDeviceCF with %d \n", del_dev));
	deleteDeviceCF(del_dev);
#endif // KERNEL_MODULE
	return 0;
}


static int config_device(int sock);	// Hilfsfunktion fr int config ()

extern struct yy_config config_dev;
extern int has_gateway, has_mask;

int config (void)
{
	int sock, on = 1;
	struct ifaddrs *ifaddr, *ifa;
	int s;
	char host[NI_MAXHOST];

	if (getifaddrs(&ifaddr) == -1) {
	  perror("getifaddrs");
	  exit(EXIT_FAILURE);
	}

	/*
	  This code loops through all interfaces, and sends an open
	  broadcast to each network interface.
	
	  Just sending a broadcast to 255.255.255.255 does not work
	  because that broadcast is only sent to the default route
	  (?).
	
	  In case of multiple configurations on a single NIC
	  (eth0:extra constructs) we are currently sending packets to
	  each configuration. It doesn't hurt, but should probably be
	  fixed.

	  The loop is based on the the demo code from man getifaddrs;
	  earlier versions of adconfig used SIOCGIFCONF, which did not
	  work properly for 64bit OSes. getifaddrs is the way to got
	  an works since glibc 2.3.3 (released in 2005, that should be
	  OK).

	  Maybe we could check for getifaddrs, and if not present: 
	  - use the old code on 32bit
	  - issue an error on 64bit
	*/
	
	for (ifa = ifaddr; ifa != NULL; ifa = ifa->ifa_next) {
	  /* ifa->ifa_addr may be NULL - the demo code in the man page
	     does not handle that case, but it occures for "tun"
	     interfaces */
	  if ( ! ifa->ifa_addr || ifa->ifa_addr->sa_family != AF_INET)
	    continue;
	  
	  s = getnameinfo(ifa->ifa_addr, sizeof(struct sockaddr_in),
			  host, NI_MAXHOST, NULL, 0, NI_NUMERICHOST);
	  if (s != 0) {
	      printf("getnameinfo() failed: %s\n", gai_strerror(s));
	      exit(EXIT_FAILURE);
	  }


	  sock = socket(AF_INET, SOCK_DGRAM, 0);
	  if (sock == -1)
	    perror("socket");
	  
	  on = 1;
	  
	  if (setsockopt(sock, SOL_SOCKET, SO_BROADCAST, &on, sizeof(on)) < 0)
	    perror("setsockopt");
	  
	  if (bind(sock, ifa->ifa_addr, sizeof(*(ifa->ifa_addr))))
	    perror("bind");
	  
	  if (config_device(sock)) return 1;
	  printf("Sent config on interface %s (%s)\n", ifa->ifa_name, host);
	  
	  close(sock);
	}
	
	freeifaddrs(ifaddr);
	return 0;
}

static int config_device (int sock)
{
	struct sockaddr_in brdcast;
	int e;
	char errstr[1000];

	if (! has_ip) {
		if (! config_dev.dhcp) {
			fprintf(stderr, "error: ip missing\n");
			return 1;
		} else
			memset(config_dev.ip, '0', sizeof(config_dev.ip));
	} else if (! has_mask) {
		if (! config_dev.dhcp) {
			fprintf(stderr, "error: mask missing\n");
			return 1;
		} else
			memset(config_dev.mask, '0', sizeof(config_dev.ip));
	} else if (! has_gateway) {
		/* Frag Thomas, Stichwort: Gerteduplizierung in Scan-tool*/
		config_dev.gateway[0] = 1;
		config_dev.gateway[1] = 0;
		config_dev.gateway[2] = 0;
		config_dev.gateway[3] = 0;
	}

	/* broadcast */
	memset(&brdcast, 0, sizeof(brdcast));
	brdcast.sin_family = AF_INET;
	brdcast.sin_port = htons (7000);
	inet_aton("255.255.255.255", &(brdcast.sin_addr));

	e = sendto(sock, &config_dev, sizeof(config_dev), 0,
		   (struct sockaddr*)&brdcast, sizeof (brdcast));
	if (e == -1) {
		sprintf(errstr, "Socket error %d (%s)\n", errno, strerror (errno));
		fprintf(stderr, errstr, errno, strerror (errno));
		return 1;
	}

	return 0;
}
